import numpy
import six
from numpy.ma import is_masked

if six.PY3:
    from typing import Sequence, Any

from mpilot.exceptions import ProgramError
from mpilot.parser.parser import CommandNode


EEMS_COMMANDS = {
    "READ": "EEMSRead",
    "CVTTOFUZZY": "CvtToFuzzy",
    "CVTTOFUZZYCURVE": "CvtToFuzzyCurve",
    "CVTTOFUZZYCAT": "CvtToFuzzyCat",
    "MEANTOMID": "MeanToMid",
    "COPYFIELD": "Copy",
    "NOT": "FuzzyNot",
    "OR": "FuzzyOr",
    "AND": "FuzzyAnd",
    "ORNEG": "FuzzyAnd",
    "XOR": "FuzzyXOr",
    "SUM": "Sum",
    "MULT": "Multiply",
    "DIVIDE": "ADividedByB",
    "MIN": "Minimum",
    "MAX": "Maximum",
    "MEAN": "Mean",
    "UNION": "FuzzyUnion",
    "DIF": "AMinusB",
    "SELECTEDUNION": "FuzzySelectedUnion",
    "WTDUNION": "FuzzyWeightedUnion",
    "WTDMEAN": "WeightedMean",
    "WTDSUM": "WeightedSum",
    "SCORERANGEBENEFIT": "ScoreRangeBenefit",
    "SCORERANGECOST": "ScoreRangeCost",
}


def flatten(li):
    # type: (Sequence[Any]) -> Sequence[Any]
    """ Flattens a list of lists of any depth to a 1D list and returns a generator """

    for item in li:
        if isinstance(item, (list, tuple)):
            for list_item in flatten(item):
                yield list_item
        else:
            yield item


def insure_fuzzy(arr, fuzzy_min, fuzzy_max):
    # type: (numpy.ma.masked_array, float, float) -> numpy.ma.masked_array
    """ Limits all array values in-place to fuzzy_min and fuzzy_max and returns the array """

    arr[arr > fuzzy_max] = fuzzy_max
    arr[arr < fuzzy_min] = fuzzy_min

    if is_masked(arr):
        arr.data[arr.mask] = arr.fill_value

    return arr


def convert_eems2_commands(command_nodes):
    # type: (Sequence[CommandNode]) -> Sequence[CommandNode]
    """ Converts command nodes returned by the parser to their MPilot equivalents """

    def find_argument(node, name):
        for argument in node.arguments:
            if argument.name == name:
                return argument.value.value

    converted = []

    for node in command_nodes:
        try:
            converted.append(
                CommandNode(
                    node.result_name
                    or find_argument(node, "NewFieldName")
                    or find_argument(node, "InFieldName"),
                    EEMS_COMMANDS.get(node.command, node.command),
                    [
                        arg
                        for arg in node.arguments
                        if arg.name not in ("NewFieldName", "OutFileName")
                    ],
                    node.lineno,
                )
            )
        except StopIteration:
            raise ProgramError(
                lineno=node.lineno,
                message="Cannot convert from EEMS 2.0: No InFieldName argument for command without a result name.",
            )

    return converted



import six
from six import python_2_unicode_compatible

from mpilot.exceptions import ProgramError

if six.PY3:
    from typing import Tuple


@python_2_unicode_compatible
class EmptyDataFile(ProgramError):
    def __init__(self, path, lineno=None):
        # type: (str, int) -> None

        super(EmptyDataFile, self).__init__(lineno)

        self.path = path

    def __str__(self):
        return "\n".join(
            (
                "Problem: The input file is empty: {}".format(self.path),
                "Solution: Double check the path and the contents of the data file.",
            )
        )


@python_2_unicode_compatible
class InvalidDataFile(ProgramError):
    def __init__(self, problem, lineno=None):
        # type: (str, int) -> None

        super(InvalidDataFile, self).__init__(lineno)

        self.problem = problem

    def __str__(self):
        return "\n".join(
            ("Problem" + self.problem, "Solution: Double check the data file.")
        )


@python_2_unicode_compatible
class MixedArrayShapes(ProgramError):
    def __init__(self, shape_a, shape_b, lineno=None):
        # type: (Tuple[int, ...], Tuple[int, ...], int) -> None

        super(MixedArrayShapes, self).__init__(lineno)

        self.shape_a = shape_a
        self.shape_b = shape_b

    def __str__(self):
        return "\n".join(
            (
                "Problem: The shapes of at least two arrays do no match. The shapes are ({}, {}) and ({}, {}).".format(
                    *(self.shape_a + self.shape_b)
                ),
                "Solution: All arrays must have the same shape. Double check the shapes of your input data.",
            )
        )


@python_2_unicode_compatible
class InvalidDirection(ProgramError):
    def __init__(self, value, lineno=None):
        # type: (str, int) -> NOne

        super(InvalidDirection, self).__init__(lineno)

        self.value = value

    def __str__(self):
        return "\n".join(
            (
                "Problem: {} is an invalid direction value.".format(self.value),
                "Solution: Use a value of LowToHigh or HighToLow for direction.",
            )
        )


@python_2_unicode_compatible
class InvalidThresholds(ProgramError):
    def __str__(self):
        return "\n".join(
            (
                "Problem: True and False thresholds must not be equal.",
                "Solution: Make sure that the TrueThreshold and FalseThreshold parameters are distinct.",
            )
        )


@python_2_unicode_compatible
class MixedArrayLengths(ProgramError):
    def __init__(self, len_a, len_b, lineno=None):
        # type: (int, int, int) -> None

        super(MixedArrayLengths, self).__init__(lineno)

        self.len_a = len_a
        self.len_b = len_b

    def __str__(self):
        return "\n".join(
            (
                "Problem: Array lengths don't match: {} and {}".format(
                    self.len_a, self.len_b
                ),
                "Solution: Make sure that the array lengths are equal.",
            )
        )


@python_2_unicode_compatible
class DuplicateRawValues(ProgramError):
    def __str__(self):
        return "\n".join(
            (
                "Problem: Raw values are not unique.",
                "Solution: Double check that the raw values are unique.",
            )
        )


@python_2_unicode_compatible
class IsNotFuzzy(ProgramError):
    def __str__(self):
        return "\n".join(
            (
                "Problem: One or more inputs are not fuzzy.",
                "Solution: Make sure the inputs to this function are fuzzy.",
            )
        )


@python_2_unicode_compatible
class InvalidNumberToConsider(ProgramError):
    def __str__(self):
        return "\n".join(
            (
                "Problem: The number of InFieldNames must be greater than or equal to NumberToConsider.",
                "Solution: Double check that the number of inputs is at least as much as NumberToConsider.",
            )
        )


@python_2_unicode_compatible
class InvalidTruestOrFalsest(ProgramError):
    def __init__(self, value, lineno=None):
        # type: (str, int) -> None

        super(InvalidTruestOrFalsest, self).__init__(lineno)

        self.value = value

    def __str__(self):
        return "\n".join(
            (
                "Problem: The value '{}' is invalid.".format(self.value),
                "Solution: Make sure that the value is either Truest or Falsest.",
            )
        )



from six import python_2_unicode_compatible

from mpilot.exceptions import MPilotError


@python_2_unicode_compatible
class NoSuchVariable(MPilotError):
    def __init__(self, path, variable, lineno=None):
        # type: (str, str, int) -> None

        super(NoSuchVariable, self).__init__(lineno)

        self.path = path
        self.variable = variable

    def __str__(self):
        return "\n".join(
            (
                "Problem: The dataset has no variable named '{}': {}".format(self.variable, self.path),
                "Solution: Double check the path and contents of the dataset.",
            )
        )


@python_2_unicode_compatible
class InvalidPositiveData(MPilotError):
    def __init__(self, path, expected_type, lineno=None):
        # type: (str, str, int) -> None

        super(InvalidPositiveData, self).__init__(lineno)

        self.path = path
        self.expected_type = expected_type

    def __str__(self):
        return "\n".join(
            (
                "Problem: The expected datatype is '{}', but the dataset has a negative value: {}".format(
                    self.expected_type, self.path
                ),
                "Solution: Double check the input data type and dataset.",
            )
        )


@python_2_unicode_compatible
class InvalidFuzzyData(MPilotError):
    def __init__(self, path, lineno=None):
        # type: (str, int) -> None

        super(InvalidPositiveData, self).__init__(lineno)

        self.path = path

    def __str__(self):
        return "\n".join(
            (
                "Problem: The expected datatype is 'Fuzzy', but the dataset has invalid values: {}".format(self.path),
                "Solution: Make sure the input dataset has values within the fuzzy range.",
            )
        )



from __future__ import absolute_import

import pkgutil
from collections import Counter, OrderedDict
from importlib import import_module

import six

if six.PY3:
    from importlib.util import module_from_spec

    from typing import Dict, Any, Union, TextIO, Sequence, Type
    from types import ModuleType

from .arguments import Argument, ListArgument
from .commands import Command
from .exceptions import (
    CommandDoesNotExist,
    DuplicateResult,
    MissingParameters,
    NoSuchParameter,
    MPilotError,
)
from .params import ResultParameter, ListParameter
from .parser.parser import Parser, ProgramNode
from .utils import flatten, EEMS_COMMANDS, convert_eems2_commands

EEMS_CSV_LIBRARIES = (
    "mpilot.libraries.eems.basic",
    "mpilot.libraries.eems.csv",
    "mpilot.libraries.eems.fuzzy",
)
EEMS_NETCDF_LIBRARIES = (
    "mpilot.libraries.eems.basic",
    "mpilot.libraries.eems.netcdf",
    "mpilot.libraries.eems.fuzzy",
)


class Program(object):
    """ A program consists of connected MPilot commands, and the arguments that will be used to run them. """

    def __init__(self, libraries=EEMS_CSV_LIBRARIES, working_dir=None):
        # type: (Sequence[str], str) -> None

        # Commands lookup, in the form of {result_name: command, ...}
        self.commands = {}

        # Load command libraries
        for lib in libraries:
            self.load_commands(lib)

        # Build a lookup of command classes by name
        library_commands = [
            info
            for info in Command.get_commands()
            if any(info.module.startswith(lib) for lib in libraries)
        ]
        duplicates = [
            name
            for name, ct in Counter(c.command.name for c in library_commands).items()
            if ct > 1
        ]
        if duplicates:
            raise MPilotError(
                "The following commands are duplicated in the libraries used in this program: {}".format(
                    ", ".join(name for name in duplicates)
                )
            )

        self.command_library = {
            info.command.name: info.command for info in library_commands
        }

        self.working_dir = working_dir

    @classmethod
    def load_commands(cls, module):
        # type: (Union[str, ModuleType]) -> None
        """
        Discovers and loads commands from the given module and any sub modules. This method is mostly used internally.
        """

        if isinstance(module, six.string_types):
            if isinstance(module, bytes):
                module = module.decode()

            module = import_module(module)

        if hasattr(module, "__path__"):
            if six.PY3:
                for info, name, _ in pkgutil.walk_packages(
                    module.__path__, prefix=module.__name__ + "."
                ):
                    spec = info.find_spec(name)
                    new_module = module_from_spec(spec)
                    spec.loader.exec_module(new_module)
            else:
                for info in pkgutil.iter_modules(module.__path__):
                    name = info[1]
                    new_module = import_module("{}.{}".format(module.__name__, name))
                    cls.load_commands(new_module)

    @classmethod
    def from_source(cls, source, libraries=EEMS_CSV_LIBRARIES, working_dir=None):
        # type: (str, Sequence[str], str) -> Program
        """ Creates a program from MPilot source code """

        def resolve_list(name, expression_node):
            """ Recursively resolves parsed list expressions into ListArgument values. """

            return ListArgument(
                name,
                [
                    resolve_list(name, n) if isinstance(n.value, list) else n.value
                    for n in expression_node.value
                ],
                lineno=expression_node.lineno,
                list_linenos=[n.lineno for n in expression_node.value],
            )

        program = cls(libraries=libraries, working_dir=working_dir)
        program_node = Parser().parse(source)

        if program_node.version == 2 or any(
            node.command in EEMS_COMMANDS for node in program_node.commands
        ):
            program_node = ProgramNode(convert_eems2_commands(program_node.commands), 3)

        for node in program_node.commands:
            command_cls = program.find_command_class(node.command)
            if not command_cls:
                raise CommandDoesNotExist(node.command, node.lineno)

            arguments = OrderedDict()
            for argument_node in node.arguments:
                if isinstance(argument_node.value.value, list):
                    arguments[argument_node.name] = resolve_list(
                        argument_node.name, argument_node.value
                    )
                elif isinstance(argument_node.value.value, dict):
                    arguments[argument_node.name] = Argument(
                        argument_node.name,
                        {k: v.value for k, v in argument_node.value.value.items()},
                        argument_node.lineno,
                    )
                else:
                    arguments[argument_node.name] = Argument(
                        argument_node.name,
                        argument_node.value.value,
                        argument_node.lineno,
                    )

            program.add_command(command_cls, node.result_name, arguments, node.lineno)

        return program

    def find_command_class(self, name):
        # type: (str) -> Type[Command]
        """ Looks up and returns a command class by name, or returns None if the command doesn't exist. """

        return self.command_library.get(name)

    def add_command(self, command_cls, result_name, arguments, lineno=None):
        # type: (Type[Command], str, Dict[str, Any], int) -> None
        """ Adds a command to the program. """

        if result_name in self.commands:
            raise DuplicateResult(result_name, lineno=lineno)

        missing_params = set(command_cls.required_inputs.keys()).difference(
            arguments.keys()
        )
        if missing_params:
            raise MissingParameters(command_cls, missing_params, lineno=lineno)

        command_args = []
        for (name, value) in arguments.items():
            if name not in command_cls.inputs and not command_cls.allow_extra_inputs:
                raise NoSuchParameter(
                    command_cls,
                    name,
                    lineno=value.lineno if isinstance(value, Argument) else None,
                )

            if isinstance(value, Argument):
                command_args.append(value)
            else:
                command_args.append(Argument(name, value))

        self.commands[result_name] = command_cls(
            result_name, command_args, program=self, lineno=lineno
        )

    def to_string(self):
        # type: () -> str
        """ Returns a string with commands formatted in the MPilot command file syntax. """

        def serialize_value(value, argument, command):
            # type: (Any, Argument, Command) -> str

            param = command.inputs[argument.name]

            if isinstance(param, ResultParameter) or (
                isinstance(param, ListParameter)
                and isinstance(param.value_type, ResultParameter)
            ):
                return str(value)
            if isinstance(value, six.string_types):
                return '"{}"'.format(value)
            else:
                return str(value)

        def serialize_argument(argument, command):
            # type: (Argument, Command) -> str

            if isinstance(argument, ListArgument):
                return "[{}]".format(
                    ", ".join(
                        serialize_value(x, argument, command) for x in argument.value
                    )
                )
            elif isinstance(argument.value, dict):
                return "[\n{}\n    ]".format(
                    ",\n".join(
                        '        "{}": "{}"'.format(key, value)
                        for key, value in argument.value.items()
                    )
                )
            return serialize_value(argument.value, argument, command)

        def serialize_command(command):
            # type: (Command) -> str

            return "{} = {}({})".format(
                command.result_name,
                command.name,
                "\n    {}\n".format(
                    ",\n    ".join(
                        "{} = {}".format(a.name, serialize_argument(a, command))
                        for a in command.arguments
                    )
                ),
            )

        return "\n".join(
            serialize_command(command) for command in self.commands.values()
        )

    def to_file(self, file_or_path):
        # type: (Union[TextIO, str]) -> None
        """ Writes the program as an MPilot command file. """

        if hasattr(file_or_path, "write"):
            f = file_or_path
        else:
            f = open(file_or_path, "w")

        f.write(self.to_string())

    def run(self):
        # Build dependency lookup
        dependents = {}  # {result_name, [dependent_name, ...], ...}

        for command in self.commands.values():
            references = []
            for argument in command.arguments:
                if argument.name in command.inputs:
                    value = command.inputs[argument.name].clean(
                        argument.value, self, lineno=argument.lineno
                    )

                    if isinstance(command.inputs[argument.name], ResultParameter):
                        references.append(
                            value.result_name if isinstance(value, Command) else value
                        )
                    if isinstance(value, (list, tuple)):
                        references += [
                            x for x in flatten(value) if isinstance(x, Command)
                        ]

            for reference in references:
                dependents[reference] = dependents.get(reference, set())
                dependents[reference].add(command.result_name)

        # Find and run leaf nodes (commands without any dependents)
        for command in (
            command
            for command in self.commands.values()
            if not dependents.get(command.result_name)
        ):
            command.run()



from __future__ import division

from functools import reduce

import numpy
import six
from packaging import version

if six.PY3:
    from typing import Union, Sequence

from mpilot import params
from mpilot.commands import Command
from mpilot.libraries.eems.exceptions import (
    InvalidDirection,
    InvalidThresholds,
    MixedArrayLengths,
    DuplicateRawValues,
    IsNotFuzzy,
    InvalidNumberToConsider,
    InvalidTruestOrFalsest,
)
from mpilot.libraries.eems.mixins import SameArrayShapeMixin
from mpilot.utils import insure_fuzzy

FUZZY_MIN = -1
FUZZY_MAX = 1


class FuzzyCommand(Command):
    is_fuzzy = False

    def validate_fuzzy_inputs(self, values, lineno=None):
        # type: (Union[Sequence[Command], Command], int) -> None

        if not isinstance(values, (list, tuple)):
            values = [values]

        for cmd in values:
            if not isinstance(cmd, FuzzyCommand) or not cmd.is_fuzzy:
                raise IsNotFuzzy(lineno=lineno)


class CvtToFuzzy(FuzzyCommand):
    """ Converts input values into fuzzy values using linear interpolation """

    is_fuzzy = True

    display_name = "Convert to Fuzzy"
    inputs = {
        "InFieldName": params.ResultParameter(params.DataParameter()),
        "TrueThreshold": params.NumberParameter(required=False),
        "FalseThreshold": params.NumberParameter(required=False),
        "Direction": params.StringParameter(required=False),
    }
    output = params.DataParameter()

    def execute(self, **kwargs):
        arr = kwargs["InFieldName"].result
        direction = kwargs.get("Direction")

        if direction and direction not in ("LowToHigh", "HighToLow"):
            raise InvalidDirection(
                direction, lineno=self.argument_lines.get("Direction")
            )

        false_threshold = kwargs.get(
            "FalseThreshold", arr.max() if direction == "HighToLow" else arr.min()
        )
        true_threshold = kwargs.get(
            "TrueThreshold", arr.min() if direction == "HighToLow" else arr.max()
        )

        if true_threshold == false_threshold:
            raise InvalidThresholds(self.lineno)

        x1 = float(true_threshold)
        x2 = float(false_threshold)
        y1 = FUZZY_MAX
        y2 = FUZZY_MIN

        result = arr - x1
        result *= y2 - y1
        result /= x2 - x1
        result += y1

        return insure_fuzzy(result, FUZZY_MIN, FUZZY_MAX)


class CvtToFuzzyZScore(FuzzyCommand):
    """ Converts input values into fuzzy values using linear interpolation based on Z Score """

    is_fuzzy = True

    display_name = "Convert to Fuzzy by Z Score"
    inputs = {
        "InFieldName": params.ResultParameter(params.DataParameter()),
        "TrueThresholdZScore": params.NumberParameter(),
        "FalseThresholdZScore": params.NumberParameter(),
    }
    output = params.DataParameter()

    def execute(self, **kwargs):
        arr = kwargs["InFieldName"].result
        true_threshold = float(kwargs["TrueThresholdZScore"])
        false_threshold = float(kwargs["FalseThresholdZScore"])

        raw_mean = numpy.ma.mean(arr)
        raw_std = numpy.ma.std(arr)

        x1 = raw_mean + raw_std * true_threshold
        x2 = raw_mean + raw_std * false_threshold
        y1 = FUZZY_MAX
        y2 = FUZZY_MIN

        result = arr.copy()
        result -= x1
        result *= y2 - y1
        result /= x2 - x1
        result += y1

        return insure_fuzzy(result, FUZZY_MIN, FUZZY_MAX)


class CvtToFuzzyCat(FuzzyCommand):
    """ Converts integer input values into fuzzy based on user specification """

    is_fuzzy = True

    display_name = "Convert to Fuzzy by Category"
    inputs = {
        "InFieldName": params.ResultParameter(params.DataParameter()),
        "RawValues": params.ListParameter(params.NumberParameter()),
        "FuzzyValues": params.ListParameter(params.NumberParameter()),
        "DefaultFuzzyValue": params.NumberParameter(),
    }
    output = params.DataParameter()

    def execute(self, **kwargs):
        arr = kwargs["InFieldName"].result
        raw_values = kwargs["RawValues"]
        fuzzy_values = kwargs["FuzzyValues"]
        default_fuzzy_value = kwargs["DefaultFuzzyValue"]

        if len(raw_values) != len(fuzzy_values):
            raise MixedArrayLengths(
                len(raw_values), len(fuzzy_values), lineno=self.lineno
            )

        if len(raw_values) != len(set(raw_values)):
            raise DuplicateRawValues(lineno=self.argument_lines.get("RawValues"))

        result = numpy.ma.array(numpy.full(arr.shape, default_fuzzy_value, dtype=float))

        for raw, fuzzy in zip(raw_values, fuzzy_values):
            result[arr.data == raw] = fuzzy

        return insure_fuzzy(result, FUZZY_MIN, FUZZY_MAX)


class CvtToFuzzyCurve(FuzzyCommand):
    """ Converts input values into fuzzy based on user-defined curve """

    is_fuzzy = True

    display_name = "Convert to Fuzzy Curve"
    inputs = {
        "InFieldName": params.ResultParameter(params.DataParameter()),
        "RawValues": params.ListParameter(params.NumberParameter()),
        "FuzzyValues": params.ListParameter(params.NumberParameter()),
    }
    output = params.DataParameter()

    def execute(self, **kwargs):
        arr = kwargs["InFieldName"].result
        raw_values = kwargs["RawValues"]
        fuzzy_values = kwargs["FuzzyValues"]

        if len(raw_values) != len(fuzzy_values):
            raise MixedArrayLengths(
                len(raw_values), len(fuzzy_values), lineno=self.lineno
            )

        if len(raw_values) != len(set(raw_values)):
            raise DuplicateRawValues(lineno=self.argument_lines.get("RawValues"))

        result = numpy.ma.empty(arr.shape, dtype=float)
        value_pairs = sorted(zip(raw_values, fuzzy_values))

        # For raw values less than the lowest raw value, set them to the corresponding fuzzy value
        result[arr <= value_pairs[0][0]] = value_pairs[0][1]

        # Assign fuzzy values for each of the line segments that approximate the curve
        for i, (raw, fuzzy) in list(enumerate(value_pairs))[1:]:
            prev_raw = value_pairs[i - 1][0]
            prev_fuzzy = value_pairs[i - 1][1]

            m = (fuzzy - prev_fuzzy) / (raw - prev_raw)
            b = prev_fuzzy - m * prev_raw

            where_idx = numpy.where(
                numpy.logical_and(arr.data > prev_raw, arr.data <= raw)
            )

            result[where_idx] = arr.data[where_idx]
            result[where_idx] *= m
            result[where_idx] += b

        # For raw values greater than the highest raw value, set them to the corresponding fuzzy value
        result[arr > value_pairs[-1][0]] = value_pairs[-1][1]
        result.mask = arr.mask.copy()

        return insure_fuzzy(result, FUZZY_MIN, FUZZY_MAX)


class CvtToFuzzyMeanToMid(CvtToFuzzyCurve):
    """ Uses "CvtToFuzzyCurve" to create a non-linear transformation that is a good match for the input data """

    is_fuzzy = True

    display_name = "Mean to Mid"
    inputs = {
        "InFieldName": params.ResultParameter(params.DataParameter()),
        "IgnoreZeros": params.BooleanParameter(),
        "FuzzyValues": params.ListParameter(params.NumberParameter()),
    }
    output = params.DataParameter()

    def execute(self, **kwargs):
        arr = kwargs["InFieldName"].result
        ignore_zeros = kwargs["IgnoreZeros"]

        low_value = arr.min()
        high_value = arr.max()

        if ignore_zeros:
            arr = arr[arr != 0]

        mean_value = arr.mean()
        below_mean = arr[arr <= mean_value]
        above_mean = arr[arr > mean_value]

        high_mean = above_mean.compressed().mean()
        low_mean = below_mean.compressed().mean()

        return super(CvtToFuzzyMeanToMid, self).execute(
            InFieldName=kwargs["InFieldName"],
            RawValues=[low_value, low_mean, mean_value, high_mean, high_value],
            FuzzyValues=kwargs["FuzzyValues"],
        )


class CvtToFuzzyCurveZScore(FuzzyCommand):
    """ Converts input values into fuzzy based on user-defined curve """

    is_fuzzy = True

    display_name = "Convert to Fuzzy Curve by Z Score"
    inputs = {
        "InFieldName": params.ResultParameter(params.DataParameter()),
        "ZScoreValues": params.ListParameter(params.NumberParameter()),
        "FuzzyValues": params.ListParameter(params.NumberParameter()),
    }
    output = params.DataParameter()

    def execute(self, **kwargs):
        arr = kwargs["InFieldName"].result
        z_score_values = kwargs["ZScoreValues"]
        fuzzy_values = kwargs["FuzzyValues"]

        if len(z_score_values) != len(fuzzy_values):
            raise MixedArrayLengths(
                len(z_score_values), len(fuzzy_values), lineno=self.lineno
            )

        raw_mean = numpy.ma.mean(arr)
        raw_std = numpy.ma.std(arr)

        raw_values = [raw_mean + value * raw_std for value in z_score_values]

        result = numpy.ma.empty(arr.shape, dtype=float)
        value_pairs = sorted(zip(raw_values, fuzzy_values))

        # For raw values less than the lowest raw value, set them to the corresponding fuzzy value
        result[arr <= value_pairs[0][0]] = value_pairs[0][1]

        # Assign fuzzy values for each of the line segments that approximate the curve
        for i, (raw, fuzzy) in list(enumerate(value_pairs))[1:]:
            prev_raw = value_pairs[i - 1][0]
            prev_fuzzy = value_pairs[i - 1][1]

            m = (fuzzy - prev_fuzzy) / (raw - prev_raw)
            b = prev_fuzzy - m * prev_raw

            where_idx = numpy.where(
                numpy.logical_and(arr.data > prev_raw, arr.data <= raw)
            )
            result[where_idx] = arr.data[where_idx]
            result[where_idx] *= m
            result[where_idx] += b

        # For raw values greater than the highest raw value, set them to the corresponding fuzzy value
        result[arr > value_pairs[-1][0]] = value_pairs[-1][1]
        result.mask = arr.mask.copy()

        return insure_fuzzy(result, FUZZY_MIN, FUZZY_MAX)


class CvtToBinary(FuzzyCommand):
    """
    Converts input values into binary 0 or 1 based on threshold.
    Direction = LowToHigh for values below threshold to be false and above to be true.
    Direction = HighToLow for values below threshold to be true and above to be false.
    """

    is_fuzzy = True

    display_name = "Convert to Fuzzy Binary"
    inputs = {
        "InFieldName": params.ResultParameter(params.DataParameter()),
        "Threshold": params.NumberParameter(),
        "Direction": params.StringParameter(),
    }
    output = params.DataParameter()

    def execute(self, **kwargs):
        arr = kwargs["InFieldName"].result
        threshold = kwargs["Threshold"]
        direction = kwargs["Direction"]

        if direction not in ("LowToHigh", "HighToLow"):
            raise InvalidDirection(
                direction, lineno=self.argument_lines.get("Direction")
            )

        low_value = 0.0 if direction == "LowToHigh" else 1.0
        high_value = 1.0 if direction == "LowToHigh" else 0.0

        result = numpy.ma.where(arr < threshold, low_value, high_value)

        return insure_fuzzy(result, FUZZY_MIN, FUZZY_MAX)


class FuzzyUnion(SameArrayShapeMixin, FuzzyCommand):
    """ Takes the fuzzy Union (mean) of fuzzy input variables """

    is_fuzzy = True

    display_name = "Fuzzy Union"
    inputs = {
        "InFieldNames": params.ListParameter(
            params.ResultParameter(params.DataParameter())
        )
    }
    output = params.DataParameter()

    def execute(self, **kwargs):
        arrays = [c.result for c in kwargs["InFieldNames"]]

        self.validate_fuzzy_inputs(
            kwargs["InFieldNames"], lineno=self.argument_lines.get("InFieldNames")
        )
        self.validate_array_shapes(
            arrays, lineno=self.argument_lines.get("InFieldNames")
        )

        result = sum(arrays)
        result /= float(len(arrays))

        return insure_fuzzy(result, FUZZY_MIN, FUZZY_MAX)


class FuzzyWeightedUnion(SameArrayShapeMixin, FuzzyCommand):
    """ Takes the weighted fuzzy Union (mean) of fuzzy input variables """

    is_fuzzy = True

    display_name = "Fuzzy Weighted Union"
    inputs = {
        "InFieldNames": params.ListParameter(
            params.ResultParameter(params.DataParameter())
        ),
        "Weights": params.ListParameter(params.NumberParameter()),
    }
    output = params.DataParameter()

    def execute(self, **kwargs):
        arrays = [c.result for c in kwargs["InFieldNames"]]
        weights = kwargs["Weights"]

        self.validate_fuzzy_inputs(
            kwargs["InFieldNames"], lineno=self.argument_lines.get("InFieldNames")
        )
        self.validate_array_shapes(
            arrays, lineno=self.argument_lines.get("InFieldNames")
        )

        result = arrays[0] * weights[0]
        for weight, arr in zip(weights[1:], arrays[1:]):
            result += arr * weight

        result /= sum(weights)

        return insure_fuzzy(result, FUZZY_MIN, FUZZY_MAX)


class FuzzySelectedUnion(SameArrayShapeMixin, FuzzyCommand):
    """ Takes the fuzzy Union (mean) of N Truest or Falsest fuzzy input variables """

    is_fuzzy = True

    display_name = "Fuzzy Selected Union"
    inputs = {
        "InFieldNames": params.ListParameter(
            params.ResultParameter(params.DataParameter())
        ),
        "TruestOrFalsest": params.StringParameter(),
        "NumberToConsider": params.NumberParameter(),
    }
    output = params.DataParameter()

    def execute(self, **kwargs):
        for cmd in kwargs["InFieldNames"]:
            if not isinstance(cmd, FuzzyCommand):
                raise IsNotFuzzy(lineno=self.argument_lines.get("InFieldNames"))

        arrays = [c.result for c in kwargs["InFieldNames"]]
        truest_or_falsest = kwargs["TruestOrFalsest"]
        number_to_consider = kwargs["NumberToConsider"]

        self.validate_array_shapes(
            arrays, lineno=self.argument_lines.get("InFieldNames")
        )
        self.validate_fuzzy_inputs(
            kwargs["InFieldNames"], lineno=self.argument_lines.get("InFieldNames")
        )
        if len(arrays) < number_to_consider:
            raise InvalidNumberToConsider(
                lineno=self.argument_lines.get("NumberToConsider")
            )

        if truest_or_falsest not in ("Truest", "Falsest"):
            raise InvalidTruestOrFalsest(
                truest_or_falsest, lineno=self.argument_lines.get("TruestOrFalsest")
            )

        # Create a stacked array with layers from input arrays, sort, and use to calculate fuzzy xor. Since there is no
        # numpy.ma.stacked, the masks are handled separately from the data. Note: we are building the maximal mask from
        # all the inputs before broadcasting it to the size of the stacked array.

        mask = reduce(
            lambda x, y: numpy.logical_or(x, y),
            (arr.mask for arr in arrays[1:]),
            arrays[0].mask,
        )

        if version.parse(numpy.__version__) >= version.parse("1.10"):
            stacked_mask = numpy.broadcast_to(
                mask, [len(arrays)] + list(arrays[0].shape)
            )
        else:
            it = numpy.nditer(
                mask,
                flags=["multi_index", "refs_ok", "zerosize_ok"],
                itershape=[len(arrays)] + list(arrays[0].shape),
                order="C",
            )
            stacked_mask = it.itviews[0]

        stacked_arr = numpy.ma.array(
            numpy.vstack([arr.data for arr in arrays]),
            mask=stacked_mask.copy()  # The array is un-writable (`.sort` will fail) without `.copy()`
        )

        stacked_arr.sort(axis=0, kind="heapsort")

        if truest_or_falsest == "Truest":
            result = numpy.ma.mean(stacked_arr[-number_to_consider:], axis=0)
        else:
            result = numpy.ma.mean(stacked_arr[:number_to_consider], axis=0)

        return insure_fuzzy(result, FUZZY_MIN, FUZZY_MAX)


class FuzzyOr(SameArrayShapeMixin, FuzzyCommand):
    """ Takes the fuzzy Or (maximum) of fuzzy input variables """

    is_fuzzy = True

    display_name = "Fuzzy Or"
    inputs = {
        "InFieldNames": params.ListParameter(
            params.ResultParameter(params.DataParameter())
        )
    }
    output = params.DataParameter()

    def execute(self, **kwargs):
        self.validate_fuzzy_inputs(
            kwargs["InFieldNames"], lineno=self.argument_lines.get("InFieldNames")
        )
        arrays = [c.result for c in kwargs["InFieldNames"]]
        self.validate_array_shapes(
            arrays, lineno=self.argument_lines.get("InFieldNames")
        )

        result = reduce(lambda x, y: numpy.ma.maximum(x, y), arrays[1:], arrays[0])

        return insure_fuzzy(result, FUZZY_MIN, FUZZY_MAX)


class FuzzyAnd(SameArrayShapeMixin, FuzzyCommand):
    """ Takes the fuzzy And (minimum) of fuzzy input variables """

    is_fuzzy = True

    display_name = "Fuzzy And"
    inputs = {
        "InFieldNames": params.ListParameter(
            params.ResultParameter(params.DataParameter())
        )
    }
    output = params.DataParameter()

    def execute(self, **kwargs):
        self.validate_fuzzy_inputs(
            kwargs["InFieldNames"], lineno=self.argument_lines.get("InFieldNames")
        )
        arrays = [c.result for c in kwargs["InFieldNames"]]
        self.validate_array_shapes(
            arrays, lineno=self.argument_lines.get("InFieldNames")
        )

        result = reduce(lambda x, y: numpy.ma.minimum(x, y), arrays[1:], arrays[0])

        return insure_fuzzy(result, FUZZY_MIN, FUZZY_MAX)


class FuzzyXOr(SameArrayShapeMixin, FuzzyCommand):
    """ Computes Fuzzy XOr: Truest - (Truest - 2nd Truest) * (2nd Truest - full False)/(Truest - full False) """

    is_fuzzy = True

    display_name = "Fuzzy XOr"
    inputs = {
        "InFieldNames": params.ListParameter(
            params.ResultParameter(params.DataParameter())
        )
    }
    output = params.DataParameter()

    def execute(self, **kwargs):
        self.validate_fuzzy_inputs(
            kwargs["InFieldNames"], lineno=self.argument_lines.get("InFieldNames")
        )
        arrays = [c.result for c in kwargs["InFieldNames"]]
        self.validate_array_shapes(
            arrays, lineno=self.argument_lines.get("InFieldNames")
        )

        # Create a stacked array with layers from input arrays, sort, and use to calculate fuzzy xor. Since there is no
        # numpy.ma.stacked, the masks are handled separately from the data. Note: we are building the maximal mask from
        # all the inputs before broadcasting it to the size of the stacked array.

        mask = reduce(
            lambda x, y: numpy.logical_or(x, y),
            (arr.mask for arr in arrays[1:]),
            arrays[0].mask,
        )

        if version.parse(numpy.__version__) >= version.parse("1.10"):
            stacked_mask = numpy.broadcast_to(
                mask, [len(arrays)] + list(arrays[0].shape)
            )
        else:
            it = numpy.nditer(
                mask,
                flags=["multi_index", "refs_ok", "zerosize_ok"],
                itershape=[len(arrays)] + list(arrays[0].shape),
                order="C",
            )
            stacked_mask = it.itviews[0]

        stacked_arr = numpy.ma.array(
            numpy.vstack([arr.data for arr in arrays]),
            mask=stacked_mask.copy(),
        )

        stacked_arr.sort(axis=0, kind="heapsort")

        result = numpy.ma.where(
            stacked_arr[-1] <= FUZZY_MIN,
            FUZZY_MIN,
            stacked_arr[-1]
            - (stacked_arr[-1] - stacked_arr[-2])
            * (stacked_arr[-2] - FUZZY_MIN)
            / (stacked_arr[-1] - FUZZY_MIN),
        )

        return insure_fuzzy(result, FUZZY_MIN, FUZZY_MAX)


class FuzzyNot(FuzzyCommand):
    """ Takes the fuzzy And (minimum) of fuzzy input variables """

    is_fuzzy = True

    display_name = "Fuzzy Not"
    inputs = {"InFieldName": params.ResultParameter(params.DataParameter())}
    output = params.DataParameter()

    def execute(self, **kwargs):
        self.validate_fuzzy_inputs(
            kwargs["InFieldName"], lineno=self.argument_lines.get("InFieldNames")
        )
        arr = kwargs["InFieldName"].result

        result = -arr

        return insure_fuzzy(result, FUZZY_MIN, FUZZY_MAX)


class CvtFromFuzzy(FuzzyCommand):
    """ Converts input fuzzy values into non-fuzzy values using linear interpolation """

    is_fuzzy = False

    display_name = "Convert from Fuzzy"
    inputs = {
        "InFieldName": params.ResultParameter(params.DataParameter()),
        "TrueThreshold": params.NumberParameter(),
        "FalseThreshold": params.NumberParameter(),
    }
    output = params.DataParameter()

    def execute(self, **kwargs):
        self.validate_fuzzy_inputs(
            kwargs["InFieldName"], lineno=self.argument_lines.get("InFieldName")
        )
        arr = kwargs["InFieldName"].result
        true_threshold = kwargs["TrueThreshold"]
        false_threshold = kwargs["FalseThreshold"]

        if true_threshold == false_threshold:
            raise InvalidThresholds(self.lineno)

        y1 = float(true_threshold)
        y2 = float(false_threshold)
        x1 = FUZZY_MAX
        x2 = FUZZY_MIN

        result = arr - x1
        result *= y2 - y1
        result /= x2 - x1
        result += y1

        return result



from __future__ import division

from functools import reduce

import numpy

from mpilot import params
from mpilot.commands import Command
from mpilot.libraries.eems.mixins import SameArrayShapeMixin


class Copy(Command):
    """ Copies the data from another field """

    display_name = "Copy"
    inputs = {"InFieldName": params.ResultParameter(params.DataParameter())}
    output = params.DataParameter()

    def execute(self, **kwargs):
        return numpy.copy(kwargs["InFieldName"].result)


class AMinusB(SameArrayShapeMixin, Command):
    """ Performs A - B """

    display_name = "A Minus B"
    inputs = {
        "A": params.ResultParameter(params.DataParameter()),
        "B": params.ResultParameter(params.DataParameter()),
    }
    output = params.DataParameter()

    def execute(self, **kwargs):
        a = kwargs["A"].result
        b = kwargs["B"].result
        self.validate_array_shapes([a, b], lineno=self.lineno)

        return a - b


class Sum(SameArrayShapeMixin, Command):
    """ Sums input variables """

    display_name = "Sum"
    inputs = {
        "InFieldNames": params.ListParameter(
            params.ResultParameter(params.DataParameter())
        )
    }
    output = params.DataParameter()

    def execute(self, **kwargs):
        arrays = [c.result for c in kwargs["InFieldNames"]]
        self.validate_array_shapes(arrays, lineno=self.lineno)

        result = numpy.copy(arrays[0])
        for arr in arrays[1:]:
            result += arr

        return result


class WeightedSum(SameArrayShapeMixin, Command):
    """ Takes the weighted sum of input variables """

    display_name = "Weighted Sum"
    inputs = {
        "InFieldNames": params.ListParameter(
            params.ResultParameter(params.DataParameter())
        ),
        "Weights": params.ListParameter(params.NumberParameter()),
    }
    output = params.DataParameter()

    def execute(self, **kwargs):
        weights = kwargs["Weights"]
        arrays = [c.result for c in kwargs["InFieldNames"]]
        self.validate_array_shapes(arrays, lineno=self.lineno)

        result = arrays[0] * weights[0]
        for weight, arr in zip(weights[1:], arrays[1:]):
            result += arr * weight

        return result


class Multiply(SameArrayShapeMixin, Command):
    """ Multiplies input variables """

    display_name = "Multiply"
    inputs = {
        "InFieldNames": params.ListParameter(
            params.ResultParameter(params.DataParameter())
        )
    }
    output = params.DataParameter()

    def execute(self, **kwargs):
        arrays = [c.result for c in kwargs["InFieldNames"]]
        self.validate_array_shapes(arrays, lineno=self.lineno)

        result = numpy.copy(arrays[0])
        for arr in arrays[1:]:
            result *= arr

        return result


class ADividedByB(SameArrayShapeMixin, Command):
    """ Performs A / B """

    display_name = "A Divided By B"
    inputs = {
        "A": params.ResultParameter(params.DataParameter()),
        "B": params.ResultParameter(params.DataParameter()),
    }
    output = params.DataParameter()

    def execute(self, **kwargs):
        a = kwargs["A"].result
        b = kwargs["B"].result
        self.validate_array_shapes([a, b], lineno=self.lineno)

        return a / b


class Minimum(SameArrayShapeMixin, Command):
    """ Takes the minimum input variables """

    display_name = "Minimum"
    inputs = {
        "InFieldNames": params.ListParameter(
            params.ResultParameter(params.DataParameter())
        )
    }
    output = params.DataParameter()

    def execute(self, **kwargs):
        arrays = [c.result for c in kwargs["InFieldNames"]]
        self.validate_array_shapes(arrays, lineno=self.lineno)

        return reduce(lambda x, y: numpy.ma.minimum(x, y), arrays)


class Maximum(SameArrayShapeMixin, Command):
    """ Takes the maximum input variables """

    display_name = "Maximum"
    inputs = {
        "InFieldNames": params.ListParameter(
            params.ResultParameter(params.DataParameter())
        )
    }
    output = params.DataParameter()

    def execute(self, **kwargs):
        arrays = [c.result for c in kwargs["InFieldNames"]]
        self.validate_array_shapes(arrays, lineno=self.lineno)

        return reduce(lambda x, y: numpy.ma.maximum(x, y), arrays)


class Mean(SameArrayShapeMixin, Command):
    """ Mean of input variables """

    display_name = "Mean"
    inputs = {
        "InFieldNames": params.ListParameter(
            params.ResultParameter(params.DataParameter())
        )
    }
    output = params.DataParameter()

    def execute(self, **kwargs):
        arrays = [c.result for c in kwargs["InFieldNames"]]
        self.validate_array_shapes(arrays, lineno=self.lineno)

        return sum(arrays) / len(arrays)


class WeightedMean(SameArrayShapeMixin, Command):
    """ Takes the weighted mean of input variables """

    display_name = "Weighted Mean"
    inputs = {
        "InFieldNames": params.ListParameter(
            params.ResultParameter(params.DataParameter())
        ),
        "Weights": params.ListParameter(params.NumberParameter()),
    }
    output = params.DataParameter()

    def execute(self, **kwargs):
        weights = kwargs["Weights"]
        arrays = [c.result for c in kwargs["InFieldNames"]]
        self.validate_array_shapes(arrays, lineno=self.lineno)

        result = arrays[0] * weights[0]
        for weight, arr in zip(weights[1:], arrays[1:]):
            result += arr * weight

        return result / sum(weights)


class Normalize(Command):
    """ Normalizes the data from another field to range (default 0:1) """

    display_name = "Normalize"
    inputs = {
        "InFieldName": params.ResultParameter(params.DataParameter()),
        "StartVal": params.NumberParameter(required=False),
        "EndVal": params.NumberParameter(required=False),
    }
    output = params.DataParameter()

    def execute(self, **kwargs):
        arr = kwargs["InFieldName"].result
        start = kwargs.get("StartVal", 0)
        end = kwargs.get("EndVal", 1)

        arr_min = arr.min()
        arr_max = arr.max()

        return (arr - arr_min) * (start - end) / (arr_min - arr_max) + start


class PrintVars(Command):
    """ Prints each variable in a list of variable names. """

    display_name = "Print variable(s) to screen or file"
    inputs = {
        "InFieldNames": params.ListParameter(params.ResultParameter()),
        "OutFileName": params.PathParameter(required=False),
    }
    output = params.BooleanParameter()

    def execute(self, **kwargs):
        commands = kwargs["InFieldNames"]
        out_path = kwargs.get("OutFileName")

        if out_path:
            with open(out_path, "w") as f_out:
                f_out.write(
                    "\n".join(
                        "{}: {}".format(c.result_name, c.result) for c in commands
                    )
                )
        else:
            for command in kwargs["InFieldNames"]:
                print("{}: {}".format(command.result_name, command.result))

        return True



from __future__ import absolute_import

import csv

import numpy

from mpilot import params
from mpilot.commands import Command
from mpilot.libraries.eems.exceptions import EmptyDataFile, InvalidDataFile
from mpilot.libraries.eems.mixins import SameArrayShapeMixin


class EEMSRead(Command):
    """ Reads a variable from a file """

    display_name = "Read"
    inputs = {
        "InFileName": params.PathParameter(must_exist=True),
        "InFieldName": params.StringParameter(),
        "ReturnType": params.DataTypeParameter(required=False),  # Not used
        "NewFieldName": params.StringParameter(required=False),  # Not used
        "MissingVal": params.NumberParameter(required=False),
        "DataType": params.DataTypeParameter(
            required=False, valid_types={"Float": float, "Integer": int}
        ),
    }
    output = params.DataParameter()

    def execute(self, **kwargs):
        path = kwargs["InFileName"]

        with open(path, "r") as f:
            reader = csv.reader(f)

            try:
                headers = next(reader)
            except StopIteration:
                raise EmptyDataFile(path)

            field_name = kwargs["InFieldName"]
            try:
                idx = headers.index(field_name)
            except ValueError:
                raise InvalidDataFile(
                    "The data file doesn't contain the header {}: {}".format(
                        field_name, path
                    )
                )

            values = []
            for row in reader:
                if row:
                    values.append(float(row[idx]))

        fill_value = kwargs.get("MissingVal")
        data_type = kwargs.get("DataType", float)
        if fill_value is not None:
            data = numpy.ma.array(
                values, mask=False, dtype=data_type, fill_value=data_type(fill_value)
            )
            mask = numpy.ma.where(data == data_type(fill_value), True, False)

        data = numpy.ma.array(values, mask=False, dtype=data_type)
        data.soften_mask()

        if fill_value is not None:
            data.mask = mask

        return data


class EEMSWrite(SameArrayShapeMixin, Command):
    display_name = "Write"
    inputs = {
        "OutFileName": params.PathParameter(must_exist=False),
        "OutFieldNames": params.ListParameter(
            params.ResultParameter(params.DataParameter())
        ),
    }

    def execute(self, **kwargs):
        commands = kwargs["OutFieldNames"]
        arrays = [c.result for c in commands]
        self.validate_array_shapes(arrays)

        with open(kwargs["OutFileName"], "w") as f:
            writer = csv.writer(f)

            # Write headers
            writer.writerow([c.result_name for c in commands])

            # Assumption: 1d arrays
            out_arr = numpy.ma.array(arrays)
            out_arr = out_arr.transpose([1, 0])

            writer.writerows(out_arr[i, :] for i in range(out_arr.shape[0]))



import six

if six.PY3:
    from typing import Any, List


class Argument(object):
    def __init__(self, name, value, lineno=None):
        # type: (str, Any, int) -> None

        self.name = name
        self.value = value
        self.lineno = lineno


class ListArgument(Argument):
    def __init__(self, name, value, lineno=None, list_linenos=None):
        # type: (str, List[Any], int, List[int]) -> None

        super(ListArgument, self).__init__(name, value, lineno=lineno)

        self.list_linenos = list_linenos



import os
from numbers import Number

import numpy
import six

if six.PY3:
    from typing import Any

from .arguments import Argument
from .exceptions import (
    ParameterNotValid,
    PathDoesNotExist,
    ResultTypeNotValid,
    ResultDoesNotExist,
    InvalidRelativePath,
)


class Parameter(object):
    def __init__(self, required=True):
        # type: (bool) -> None

        self.required = required

    def clean(self, value, program=None, lineno=None):
        # type: (Any, Any, int) -> Any
        """ Clean and validate a raw parameter value """

        return value


class StringParameter(Parameter):
    def __init__(self, **kwargs):
        super(StringParameter, self).__init__(**kwargs)

    def clean(self, value, program=None, lineno=None):
        return six.text_type(value)

    @staticmethod
    def accepts(parameter_cls):
        return issubclass(
            parameter_cls, (StringParameter, NumberParameter, PathParameter)
        )


class NumberParameter(Parameter):
    def clean(self, value, program=None, lineno=None):
        if isinstance(value, Number):
            return value

        try:
            return int(value)
        except ValueError:
            try:
                return float(value)
            except ValueError:
                raise ParameterNotValid(value, "Number", lineno)

    @staticmethod
    def accepts(parameter_cls):
        return issubclass(parameter_cls, NumberParameter)


class BooleanParameter(Parameter):
    def clean(self, value, program=None, lineno=None):
        if isinstance(value, bool):
            return value

        if isinstance(value, int):
            return bool(value)

        if isinstance(value, six.string_types):
            if value.lower() == "true":
                return True
            elif value.lower() == "false":
                return False

            try:
                return bool(int(value))
            except ValueError:
                pass

        raise ParameterNotValid(value, "Boolean", lineno)


class PathParameter(StringParameter):
    def __init__(self, must_exist=True, **kwargs):
        super(PathParameter, self).__init__(**kwargs)
        self.must_exist = must_exist

    def clean(self, value, program=None, lineno=None):
        super(PathParameter, self).clean(value, program, lineno)

        if not os.path.isabs(value):
            if program.working_dir is None:
                raise InvalidRelativePath(value, lineno)
            value = os.path.join(program.working_dir, value)

        if self.must_exist and not os.path.exists(value):
            raise PathDoesNotExist(value, lineno)

        return value


class ResultParameter(Parameter):
    def __init__(self, output_type=None, **kwargs):
        super(ResultParameter, self).__init__(**kwargs)
        self.output_type = output_type

    def clean(self, value, program=None, lineno=None):
        from . import commands

        if isinstance(value, six.string_types):
            try:
                value = program.commands[value]
            except KeyError:
                raise ResultDoesNotExist(value, lineno=lineno)

        if not isinstance(value, commands.Command):
            raise ParameterNotValid(value, "Result", lineno)

        if self.output_type is None:
            return value

        if value.is_finished:
            self.output_type.clean(value.result, program, lineno)
            return value
        elif value.output is not None:
            if hasattr(self.output_type, "accepts"):
                is_valid = self.output_type.accepts(value.output.__class__)
            else:
                is_valid = issubclass(
                    value.output.__class__, self.output_type.__class__
                )

            if not is_valid:
                raise ResultTypeNotValid(value.result_name, lineno)

        return value


class ListParameter(Parameter):
    def __init__(self, value_type=Parameter(), **kwargs):
        super(ListParameter, self).__init__(**kwargs)

        self.value_type = value_type

    def clean(self, value, program=None, lineno=None):
        if not isinstance(value, (list, tuple)):
            raise ParameterNotValid(value, "List", lineno)

        return [
            self.value_type.clean(
                item.value if isinstance(item, Argument) else item, program, lineno
            )
            for item in value
        ]


class TupleParameter(Parameter):
    def clean(self, value, program=None, lineno=None):
        if not (value == [] or isinstance(value, dict)):
            raise ParameterNotValid(value, "Tuple", lineno)

        return (
            {six.text_type(k): six.text_type(v) for k, v in value.items()}
            if value
            else {}
        )


class DataParameter(Parameter):
    def clean(self, value, program=None, lineno=None):
        if not isinstance(value, numpy.ndarray):
            raise ParameterNotValid(value, "Data Array", lineno)

        return value


class DataTypeParameter(StringParameter):
    def __init__(self, valid_types={"Float": float, "Integer": int}, **kwargs):
        super(DataTypeParameter, self).__init__(**kwargs)

        self.valid_types = valid_types

    def clean(self, value, program=None, lineno=None):
        if value in self.valid_types.values():
            return value

        try:
            return self.valid_types[value]
        except KeyError:
            raise ParameterNotValid(
                value,
                "Data Type ({})".format(",".join(self.valid_types.keys())),
                lineno,
            )
